@@ -7,43 +7,7 @@ import sys
 
 import indexclient.client as client
 
-
-def __log_backoff_retry(details):
-    args_str = ", ".join(map(str, details["args"]))
-    kwargs_str = (
-        (", " + _print_kwargs(details["kwargs"])) if details.get("kwargs") else ""
-    )
-    func_call_log = "{}({}{})".format(
-        _print_func_name(details["target"]), args_str, kwargs_str
-    )
-    logging.warning(
-        "backoff: call {func_call} delay {wait:0.1f} seconds after {tries} tries".format(
-            func_call=func_call_log, **details
-        )
-    )
-
-
-def __log_backoff_giveup(details):
-    args_str = ", ".join(map(str, details["args"]))
-    kwargs_str = (
-        (", " + _print_kwargs(details["kwargs"])) if details.get("kwargs") else ""
-    )
-    func_call_log = "{}({}{})".format(
-        _print_func_name(details["target"]), args_str, kwargs_str
-    )
-    logging.error(
-        "backoff: gave up call {func_call} after {tries} tries; exception: {exc}".format(
-            func_call=func_call_log, exc=sys.exc_info(), **details
-        )
-    )
-
-
-# Default settings to control usage of backoff library.
-BACKOFF_SETTINGS = {
-    "on_backoff": __log_backoff_retry,
-    "on_giveup": __log_backoff_giveup,
-    "max_tries": 2,
-}
+from gen3.utils import DEFAULT_BACKOFF_SETTINGS
 
 
 class Gen3Index:
@@ -91,7 +55,7 @@ class Gen3Index:
             return False
         return response.text == "Healthy"
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_version(self):
         """
 
@@ -102,7 +66,7 @@ class Gen3Index:
         response.raise_for_status()
         return response.json()
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_stats(self):
         """
 
@@ -113,7 +77,7 @@ class Gen3Index:
         response.raise_for_status()
         return response.json()
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_all_records(self, limit=None, paginate=False, start=None):
         """
 
@@ -158,7 +122,7 @@ class Gen3Index:
 
         return all_records
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_records_on_page(self, limit=None, page=None):
         """
 
@@ -181,7 +145,26 @@ class Gen3Index:
 
         return response.json().get("records")
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
+    async def async_get_record(self, guid=None, _ssl=None):
+        """
+        Asynchronous function to request a record from indexd.
+
+        Args:
+            guid (str): record guid
+
+        Returns:
+            dict: indexd record
+        """
+        url = f"{self.client.url}/index/{guid}"
+        async with aiohttp.ClientSession() as session:
+            async with session.get(url, ssl=_ssl) as response:
+                response.raise_for_status()
+                response = await response.json()
+
+        return response
+
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     async def async_get_records_on_page(self, limit=None, page=None, _ssl=None):
         """
         Asynchronous function to request a page from indexd.
@@ -210,7 +193,7 @@ class Gen3Index:
 
         return response.get("records")
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get(self, guid, dist_resolution=True):
         """
 
@@ -231,7 +214,7 @@ class Gen3Index:
 
         return rec.to_json()
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_urls(self, size=None, hashes=None, guids=None):
         """
 
@@ -252,7 +235,7 @@ class Gen3Index:
         urls = self.client._get("urls", params=p).json()
         return [url for _, url in urls.items()]
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_record(self, guid):
         """
 
@@ -266,7 +249,16 @@ class Gen3Index:
 
         return rec.to_json()
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
+    def get_record_doc(self, guid):
+        """
+
+        Get the metadata associated with a given id
+
+        """
+        return self.client.get(guid)
+
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_with_params(self, params=None):
         """
 
@@ -285,7 +277,33 @@ class Gen3Index:
 
         return rec.to_json()
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
+    async def async_get_with_params(self, params, _ssl=None):
+        """
+
+        Return a document object corresponding to the supplied parameter
+
+        - need to include all the hashes in the request
+        - need to handle the query param `'hash': 'hash_type:hash'`
+
+        Args:
+            params (dict): params to search with
+            _ssl (None, optional): whether or not to use ssl
+
+        Returns:
+            Document: json representation of an entry in indexd
+
+        """
+        query_params = urllib.parse.urlencode(params)
+        url = f"{self.client.url}/index/?{query_params}"
+        async with aiohttp.ClientSession() as session:
+            async with session.get(url, ssl=_ssl) as response:
+                await response.raise_for_status()
+                response = await response.json()
+
+        return response
+
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_latest_version(self, guid, has_version=False):
         """
 
@@ -306,7 +324,7 @@ class Gen3Index:
 
         return rec.to_json()
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_versions(self, guid):
         """
 
@@ -326,7 +344,7 @@ class Gen3Index:
 
     ### Post Requests
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def create_record(
         self,
         hashes,
@@ -377,7 +395,7 @@ class Gen3Index:
         )
         return rec.to_json()
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def create_blank(self, uploader, file_name=None):
         """
 
@@ -385,10 +403,10 @@ class Gen3Index:
 
         Args:
             json - json in the format:
-                    {
-                        'uploader': type(string)
-                        'file_name': type(string) (optional*)
-                      }
+            {
+                'uploader': type(string)
+                'file_name': type(string) (optional*)
+            }
 
         """
         json = {"uploader": uploader, "file_name": file_name}
@@ -403,7 +421,7 @@ class Gen3Index:
 
         return self.get_record(rec["did"])
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def create_new_version(
         self,
         guid,
@@ -422,10 +440,10 @@ class Gen3Index:
 
         Add new version for the document associated to the provided uuid
 
-        - Since data content is immutable, when you want to change the
-              size or hash, a new index document with a new uuid needs to be
-              created as its new version. That uuid is returned in the did
-              field of the response. The old index document is not deleted.
+        Since data content is immutable, when you want to change the
+        size or hash, a new index document with a new uuid needs to be
+        created as its new version. That uuid is returned in the did
+        field of the response. The old index document is not deleted.
 
         Args:
             guid: (string): record id
@@ -442,10 +460,10 @@ class Gen3Index:
             authz (str): RBAC string
 
             body: json/dictionary format
-                - Metadata object that needs to be added to the store.
-                  Providing size and at least one hash is necessary and
-                  sufficient. Note: it is a good idea to add a version
-                  number
+            - Metadata object that needs to be added to the store.
+              Providing size and at least one hash is necessary and
+              sufficient. Note: it is a good idea to add a version
+              number
 
         """
         if urls is None:
@@ -478,7 +496,7 @@ class Gen3Index:
             return self.get_record(rec["did"])
         return None
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def get_records(self, dids):
         """
 
@@ -506,7 +524,7 @@ class Gen3Index:
 
     ### Put Requests
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def update_blank(self, guid, rev, hashes, size):
         """
 
@@ -535,7 +553,7 @@ class Gen3Index:
 
         return self.get_record(rec["did"])
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def update_record(
         self,
         guid,
@@ -570,14 +588,14 @@ class Gen3Index:
         }
         rec = self.client.get(guid)
         for k, v in updatable_attrs.items():
-            if v:
+            if v is not None:
                 exec(f"rec.{k} = v")
         rec.patch()
         return rec.to_json()
 
     ### Delete Requests
 
-    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_SETTINGS)
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
     def delete_record(self, guid):
         """
 
@@ -594,6 +612,44 @@ class Gen3Index:
         rec.delete()
         return rec
 
+    ### Query Requests
+
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
+    def query_urls(self, pattern):
+        """
+
+        Query all record URLs for given pattern
+
+        Args:
+            pattern (str): pattern to match against indexd urls
+
+        Returns:
+            List[records]: indexd records with urls matching pattern
+        """
+        response = self.client._get(f"/_query/urls/q?include={pattern}")
+        response.raise_for_status()
+        return response.json()
+
+    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
+    async def async_query_urls(self, pattern, _ssl=None):
+        """
+        Asynchronous function to query urls from indexd.
+
+        Args:
+            pattern (str): pattern to match against indexd urls
+
+        Returns:
+            List[records]: indexd records with urls matching pattern
+        """
+        url = f"{self.client.url}/_query/urls/q?include={pattern}"
+        async with aiohttp.ClientSession() as session:
+            logging.debug(f"request: {url}")
+            async with session.get(url, ssl=_ssl) as response:
+                response.raise_for_status()
+                response = await response.json()
+
+        return response
+
 
 def _print_func_name(function):
     return "{}.{}".format(function.__module__, function.__name__)
